__doc__ = """
The script michelanglo_protein.aprior_effect generates the dictionary that is used to say what the apriori effect are. 
Namely, what amino acid is smaller etc.

>>> from michelanglo_protein.apriori_effect import Changedex
>>> pprint(Changedex().fill().to_dict())

Also the scores from wikipedia via the class WikiTable.
"""

from collections import defaultdict


class Changedex:
    inverter = {'S': 'B', 'B': 'S', 'E': 'E', 'D': 'D',
                'F': 'R', 'R': 'F',
                'C': 'C',
                'I': 'I',
                'A': 'a', 'a': 'A',
                'P': 'H', 'H': 'P'}
    full = {'S': 'smaller',
            'B': 'bigger',
            'E': 'equally sized',
            'F': 'more flexible',
            'R': 'more rigid',
            'C': 'differently charged',
            'H': 'more hydrophobic',
            'P': 'more polar',
            'I': 'identical',
            'D': 'differently shaped',
            'a': 'from an aromatic to a non-aromatic',
            'A': 'from a non-aromatic to an aromatic'}

    hydrophobic = tuple('AVILMFYW')
    polar = tuple('CGPSTNQ')
    positive = tuple('RHK')
    negative = tuple('DE')
    aa = tuple('AVILMFYWSTNQCGPRHKDE')
    aromatic = tuple('PYWH')

    def __init__(self):
        self._data = defaultdict(set)

    def __getitem__(self, items):
        return self._data['{0}>{1}'.format(*items)]

    def __setitem__(self, items, value):
        for i in items[0]:
            for j in items[1]:
                self._data['{0}>{1}'.format(i, j)].add(value)

    def remove(self, items, value):
        for i in items[0]:
            for j in items[1]:
                self._data['{0}>{1}'.format(i, j)].discard(value)

    def to_dict(self):
        return {k: '|'.join([self.full[i] for i in self._data[k]]) for k in self._data}

    def __str__(self):
        return str(self.to_dict())

    def fill_inverse(self):
        for a in self.aa:
            for b in self.aa:
                if a != b:
                    for v in self[a, b]:
                        self[b, a] = self.inverter[v]
                else:
                    self._data[a + '>' + a] = set('I')

    def fill(self):
        for a in self.aa:
            self[a, 'G'] = 'S'  # glycine is smallest
            if a != 'G':  # glycine followed by alaine in smallness
                self[a, 'A'] = 'S'
                if a != 'A':  # then serine
                    self[a, 'P'] = 'S'
                    if a != 'P':
                        self[a, 'SC'] = 'S'
                pass
            if a in self.positive:
                pass

        self['S', 'C'] = 'E'
        self['E', 'Q'] = 'E'
        self['D', 'N'] = 'E'
        self['L', 'DN'] = 'E'
        self['I', 'V'] = 'S'
        self['Y', 'F'] = 'S'
        self['EQ', 'DN'] = 'S'
        self['L', 'V'] = 'S'
        self['FYW', 'H'] = 'S'
        self['W', 'YF'] = 'D'
        self['EQ', 'M'] = 'S'
        self['FY', 'VTLDEM'] = 'S'
        self['W', 'VTLD'] = 'S'
        self['V', 'T'] = 'E'
        self['RK', 'MED'] = 'S'
        self['R', 'K'] = 'S'

        # inverse
        self.fill_inverse()
        for a in self.aa:
            # fill empties
            for b in self.aa:
                if not self[a, b]:
                    self[a, b] = 'D'
            # fill flex
            self[a, 'P'].add('F')  # proline is most flexible.
            if a != 'P':  #followed by glycine in flex
                self[a, 'G'].add('F')
                if a not in 'PG':
                    self[a, 'S'].add('F')
            # fill charge
            for charge in (self.positive, self.negative):
                if a in charge:
                    for b in self.aa:
                        if b not in charge:
                            self[a, b] = 'C'
            # fill polar
            if a in self.polar:
                for b in self.hydrophobic:
                    self[a, b] = 'H'
            # fill aroma
            if a in self.aromatic:
                for b in set(self.aa) - set(self.aromatic):
                    self[a, b] = 'a'
            elif a not in self.aromatic:
                for b in self.aromatic:
                    self[a, b] = 'A'

        self.fill_inverse()
        return self

class WikiTable:
    grantham = '''{| class="wikitable"
|Arg
|Leu
|Pro
|Thr
|Ala
|Val
|Gly
|Ile
|Phe
|Tyr
|Cys
|His
|Gln
|Asn
|Lys
|Asp
|Glu
|Met
|Trp
|
|-
|110
|145
|74
|58
|99
|124
|56
|142
|155
|144
|112
|89
|68
|46
|121
|65
|80
|135
|177
|Ser
|-
|
|102
|103
|71
|112
|96
|125
|97
|97
|77
|180
|29
|43
|86
|26
|96
|54
|91
|101
|Arg
|-
|
|
|98
|92
|96
|32
|138
|5
|22
|36
|198
|99
|113
|153
|107
|172
|138
|15
|61
|Leu
|-
|
|
|
|38
|27
|68
|42
|95
|114
|110
|169
|77
|76
|91
|103
|108
|93
|87
|147
|Pro
|-
|
|
|
|
|58
|69
|59
|89
|103
|92
|149
|47
|42
|65
|78
|85
|65
|81
|128
|Thr
|-
|
|
|
|
|
|64
|60
|94
|113
|112
|195
|86
|91
|111
|106
|126
|107
|84
|148
|Ala
|-
|
|
|
|
|
|
|109
|29
|50
|55
|192
|84
|96
|133
|97
|152
|121
|21
|88
|Val
|-
|
|
|
|
|
|
|
|135
|153
|147
|159
|98
|87
|80
|127
|94
|98
|127
|184
|Gly
|-
|
|
|
|
|
|
|
|
|21
|33
|198
|94
|109
|149
|102
|168
|134
|10
|61
|Ile
|-
|
|
|
|
|
|
|
|
|
|22
|205
|100
|116
|158
|102
|177
|140
|28
|40
|Phe
|-
|
|
|
|
|
|
|
|
|
|
|194
|83
|99
|143
|85
|160
|122
|36
|37
|Tyr
|-
|
|
|
|
|
|
|
|
|
|
|
|174
|154
|139
|202
|154
|170
|196
|215
|Cys
|-
|
|
|
|
|
|
|
|
|
|
|
|
|24
|68
|32
|81
|40
|87
|115
|His
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|46
|53
|61
|29
|101
|130
|Gln
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|94
|23
|42
|142
|174
|Asn
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|101
|56
|95
|110
|Lys
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|45
|160
|181
|Asp
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|126
|152
|Glu
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|67
|Met
|} 
|}'''
    sneath = '''{| class="wikitable"
|
|Leu
|Ile
|Val
|Gly
|Ala
|Pro
|Gln
|Asn
|Met
|Thr
|Ser
|Cys
|Glu
|Asp
|Lys
|Arg
|Tyr
|Phe
|Trp
|-
|Ile
|5
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|-
|Val
|9
|7
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|-
|Gly
|24
|25
|19
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|-
|Ala
|15
|17
|12
|9
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|-
|Pro
|23
|24
|20
|17
|16
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|-
|Glu
|22
|24
|25
|32
|26
|33
|
|
|
|
|
|
|
|
|
|
|
|
|
|-
|Asn
|20
|23
|23
|26
|25
|31
|10
|
|
|
|
|
|
|
|
|
|
|
|
|-
|Met
|20
|22
|23
|34
|25
|31
|13
|21
|
|
|
|
|
|
|
|
|
|
|
|-
|Thr
|23
|21
|17
|20
|20
|25
|24
|19
|25
|
|
|
|
|
|
|
|
|
|
|-
|Ser
|23
|25
|20
|19
|16
|24
|21
|15
|22
|12
|
|
|
|
|
|
|
|
|
|-
|Cys
|24
|26
|21
|21
|13
|25
|22
|19
|17
|19
|13
|
|
|
|
|
|
|
|
|-
|Glu
|30
|31
|31
|37
|34
|43
|14
|19
|26
|34
|29
|33
|
|
|
|
|
|
|
|-
|Asp
|25
|28
|28
|33
|30
|40
|22
|14
|31
|29
|25
|28
|7
|
|
|
|
|
|
|-
|Lys
|23
|24
|26
|31
|26
|31
|21
|27
|24
|34
|31
|32
|26
|34
|
|
|
|
|
|-
|Arg
|33
|34
|36
|43
|37
|43
|23
|31
|28
|38
|37
|36
|31
|39
|14
|
|
|
|
|-
|Tyr
|30
|34
|36
|36
|34
|37
|29
|28
|32
|32
|29
|34
|34
|34
|34
|36
|
|
|
|-
|Phe
|19
|22
|26
|29
|26
|27
|24
|24
|24
|28
|25
|29
|35
|35
|28
|34
|13
|
|
|-
|Trp
|30
|34
|37
|39
|36
|37
|31
|32
|31
|38
|35
|37
|43
|45
|34
|36
|21
|13
|
|-
|His
|25
|28
|31
|34
|29
|36
|27
|24
|30
|34
|28
|31
|27
|35
|27
|31
|23
|18
|25
|} 
|}'''
    epstein='''{| class="wikitable"
|
|Phe
|Met
|Leu
|Ile
|Val
|Pro
|Tyr
|Trp
|Cys
|Ala
|Gly
|Ser
|Thr
|His
|Glu
|Gln
|Asp
|Asn
|Lys
|Arg
|-
|Phe
|
|0.05
|0.08
|0.08
|0.1
|0.1
|0.21
|0.25
|0.22
|0.43
|0.53
|0.81
|0.81
|0.8
|1
|1
|1
|1
|1
|1
|-
|Met
|0.1
|
|0.03
|0.03
|0.1
|0.1
|0.25
|0.32
|0.21
|0.41
|0.42
|0.8
|0.8
|0.8
|1
|1
|1
|1
|1
|1
|-
|Leu
|0.15
|0.05
|
|0
|0.03
|0.03
|0.28
|0.36
|0.2
|0.43
|0.51
|0.8
|0.8
|0.81
|1
|1
|1
|1
|1
|1.01
|-
|Ile
|0.15
|0.05
|0
|
|0.03
|0.03
|0.28
|0.36
|0.2
|0.43
|0.51
|0.8
|0.8
|0.81
|1
|1
|1
|1
|1
|1.01
|-
|Val
|0.2
|0.1
|0.05
|0.05
|
|0
|0.32
|0.4
|0.2
|0.4
|0.5
|0.8
|0.8
|0.81
|1
|1
|1
|1
|1
|1.02
|-
|Pro
|0.2
|0.1
|0.05
|0.05
|0
|
|0.32
|0.4
|0.2
|0.4
|0.5
|0.8
|0.8
|0.81
|1
|1
|1
|1
|1
|1.02
|-
|Tyr
|0.2
|0.22
|0.22
|0.22
|0.24
|0.24
|
|0.1
|0.13
|0.27
|0.36
|0.62
|0.61
|0.6
|0.8
|0.8
|0.81
|0.81
|0.8
|0.8
|-
|Trp
|0.21
|0.24
|0.25
|0.25
|0.27
|0.27
|0.05
|
|0.18
|0.3
|0.39
|0.63
|0.63
|0.61
|0.81
|0.81
|0.81
|0.81
|0.81
|0.8
|-
|Cys
|0.28
|0.22
|0.21
|0.21
|0.2
|0.2
|0.25
|0.35
|
|0.25
|0.31
|0.6
|0.6
|0.62
|0.81
|0.81
|0.8
|0.8
|0.81
|0.82
|-
|Ala
|0.5
|0.45
|0.43
|0.43
|0.41
|0.41
|0.4
|0.49
|0.22
|
|0.1
|0.4
|0.41
|0.47
|0.63
|0.63
|0.62
|0.62
|0.63
|0.67
|-
|Gly
|0.61
|0.56
|0.54
|0.54
|0.52
|0.52
|0.5
|0.58
|0.34
|0.1
|
|0.32
|0.34
|0.42
|0.56
|0.56
|0.54
|0.54
|0.56
|0.61
|-
|Ser
|0.81
|0.8
|0.8
|0.8
|0.8
|0.8
|0.62
|0.63
|0.6
|0.4
|0.3
|
|0.03
|0.1
|0.21
|0.21
|0.2
|0.2
|0.21
|0.24
|-
|Thr
|0.81
|0.8
|0.8
|0.8
|0.8
|0.8
|0.61
|0.63
|0.6
|0.4
|0.31
|0.03
|
|0.08
|0.21
|0.21
|0.2
|0.2
|0.21
|0.22
|-
|His
|0.8
|0.8
|1
|1
|0.8
|0.8
|0.6
|0.61
|0.61
|0.42
|0.34
|0.1
|0.08
|
|0.2
|0.2
|0.21
|0.21
|0.2
|0.2
|-
|Glu
|1
|1
|1
|1
|1
|1
|0.8
|0.81
|0.8
|0.61
|0.52
|0.22
|0.21
|0.2
|
|0
|0.03
|0.03
|0
|0.05
|-
|Gln
|1
|1
|1
|1
|1
|1
|0.8
|0.81
|0.8
|0.61
|0.52
|0.22
|0.21
|0.2
|0
|
|0.03
|0.03
|0
|0.05
|-
|Asp
|1
|1
|1
|1
|1
|1
|0.81
|0.81
|0.8
|0.61
|0.51
|0.21
|0.2
|0.21
|0.03
|0.03
|
|0
|0.03
|0.08
|-
|Asn
|1
|1
|1
|1
|1
|1
|0.81
|0.81
|0.8
|0.61
|0.51
|0.21
|0.2
|0.21
|0.03
|0.03
|0
|
|0.03
|0.08
|-
|Lys
|1
|1
|1
|1
|1
|1
|0.8
|0.81
|0.8
|0.61
|0.52
|0.22
|0.21
|0.2
|0
|0
|0.03
|0.03
|
|0.05
|-
|Arg
|1
|1
|1
|1
|1.01
|1.01
|0.8
|0.8
|0.81
|0.62
|0.53
|0.24
|0.22
|0.2
|0.05
|0.05
|0.08
|0.08
|0.05
|
|}
|}'''
    miyata = '''{| class="wikitable"
|Cys
|Pro
|Ala
|Gly
|Ser
|Thr
|Gln
|Glu
|Asn
|Asp
|His
|Lys
|Arg
|Val
|Leu
|Ile
|Met
|Phe
|Tyr
|Trp
|
|-
|
|1.33
|1.39
|2.22
|2.84
|1.45
|2.48
|3.26
|2.83
|3.48
|2.56
|3.27
|3.06
|0.86
|1.65
|1.63
|1.46
|2.24
|2.38
|3.34
|Cys
|-
|
|
|0.06
|0.97
|0.56
|0.87
|1.92
|2.48
|1.8
|2.4
|2.15
|2.94
|2.9
|1.79
|2.7
|2.62
|2.36
|3.17
|3.12
|4.17
|Pro
|-
|
|
|
|0.91
|0.51
|0.9
|1.92
|2.46
|1.78
|2.37
|2.17
|2.96
|2.92
|1.85
|2.76
|2.69
|2.42
|3.23
|3.18
|4.23
|Ala
|-
|
|
|
|
|0.85
|1.7
|2.48
|2.78
|1.96
|2.37
|2.78
|3.54
|3.58
|2.76
|3.67
|3.6
|3.34
|4.14
|4.08
|5.13
|Gly
|-
|
|
|
|
|
|0.89
|1.65
|2.06
|1.31
|1.87
|1.94
|2.71
|2.74
|2.15
|3.04
|2.95
|2.67
|3.45
|3.33
|4.38
|Ser
|-
|
|
|
|
|
|
|1.12
|1.83
|1.4
|2.05
|1.32
|2.1
|2.03
|1.42
|2.25
|2.14
|1.86
|2.6
|2.45
|3.5
|Thr
|-
|
|
|
|
|
|
|
|0.84
|0.99
|1.47
|0.32
|1.06
|1.13
|2.13
|2.7
|2.57
|2.3
|2.81
|2.48
|3.42
|Gln
|-
|
|
|
|
|
|
|
|
|0.85
|0.9
|0.96
|1.14
|1.45
|2.97
|3.53
|3.39
|3.13
|3.59
|3.22
|4.08
|Glu
|-
|
|
|
|
|
|
|
|
|
|0.65
|1.29
|1.84
|2.04
|2.76
|3.49
|3.37
|3.08
|3.7
|3.42
|4.39
|Asn
|-
|
|
|
|
|
|
|
|
|
|
|1.72
|2.05
|2.34
|3.4
|4.1
|3.98
|3.69
|4.27
|3.95
|4.88
|Asp
|-
|
|
|
|
|
|
|
|
|
|
|
|0.79
|0.82
|2.11
|2.59
|2.45
|2.19
|2.63
|2.27
|3.16
|His
|-
|
|
|
|
|
|
|
|
|
|
|
|
|0.4
|2.7
|2.98
|2.84
|2.63
|2.85
|2.42
|3.11
|Lys
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|2.43
|2.62
|2.49
|2.29
|2.47
|2.02
|2.72
|Arg
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|0.91
|0.85
|0.62
|1.43
|1.52
|2.51
|Val
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|0.14
|0.41
|0.63
|0.94
|1.73
|Leu
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|0.29
|0.61
|0.86
|1.72
|Ile
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|0.82
|0.93
|1.89
|Met
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|0.48
|1.11
|Phe
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|1.06
|Tyr
|-
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|
|Trp
|} 
|-
|
|}'''
    yampolsky='''
{| class="wikitable"
|
|Cys
|Ser
|Thr
|Pro
|Ala
|Gly
|Asn
|Asp
|Glu
|Gln
|His
|Arg
|Lys
|Met
|Ile
|Leu
|Val
|Phe
|Tyr
|Trp
|Ex<sub>src</sub>
|-
|Cys
|.
|258
|121
|201
|334
|288
|109
|109
|270
|383
|258
|306
|252
|169
|109
|347
|89
|349
|349
|139
|280
|-
|Ser
|373
|.
|481
|249
|490
|418
|390
|314
|343
|352
|353
|363
|275
|321
|270
|295
|358
|334
|294
|160
|351
|-
|Thr
|325
|408
|.
|164
|402
|332
|240
|190
|212
|308
|246
|299
|256
|152
|198
|271
|362
|273
|260
|66
|287
|-
|Pro
|345
|392
|286
|.
|454
|404
|352
|254
|346
|384
|369
|254
|231
|257
|204
|258
|421
|339
|298
|305
|335
|-
|Ala
|393
|384
|312
|243
|.
|387
|430
|193
|275
|320
|301
|295
|225
|549
|245
|313
|319
|305
|286
|165
|312
|-
|Gly
|267
|304
|187
|140
|369
|.
|210
|188
|206
|272
|235
|178
|219
|197
|110
|193
|208
|168
|188
|173
|228
|-
|Asn
|234
|355
|329
|275
|400
|391
|.
|208
|257
|298
|248
|252
|183
|236
|184
|233
|233
|210
|251
|120
|272
|-
|Asp
|285
|275
|245
|220
|293
|264
|201
|.
|344
|263
|298
|252
|208
|245
|299
|236
|175
|233
|227
|103
|258
|-
|Glu
|332
|355
|292
|216
|520
|407
|258
|533
|.
|341
|380
|279
|323
|219
|450
|321
|351
|342
|348
|145
|363
|-
|Gln
|383
|443
|361
|212
|499
|406
|338
|68
|439
|.
|396
|366
|354
|504
|467
|391
|603
|383
|361
|159
|386
|-
|His
|331
|365
|205
|220
|462
|370
|225
|141
|319
|301
|.
|275
|332
|315
|205
|364
|255
|328
|260
|72
|303
|-
|Arg
|225
|270
|199
|145
|459
|251
|67
|124
|250
|288
|263
|.
|306
|68
|139
|242
|189
|213
|272
|63
|259
|-
|Lys
|331
|376
|476
|252
|600
|492
|457
|465
|272
|441
|362
|440
|.
|414
|491
|301
|487
|360
|343
|218
|409
|-
|Met
|347
|353
|261
|85
|357
|218
|544
|392
|287
|394
|278
|112
|135
|.
|612
|513
|354
|330
|308
|633
|307
|-
|Ile
|362
|196
|193
|145
|326
|160
|172
|27
|197
|191
|221
|124
|121
|279
|.
|417
|494
|331
|323
|73
|252
|-
|Leu
|366
|212
|165
|146
|343
|201
|162
|112
|199
|250
|288
|185
|171
|367
|301
|.
|275
|336
|295
|152
|248
|-
|Val
|382
|326
|398
|201
|389
|269
|108
|228
|192
|280
|253
|190
|197
|562
|537
|333
|.
|207
|209
|286
|277
|-
|Phe
|176
|152
|257
|112
|236
|94
|136
|90
|62
|216
|237
|122
|85
|255
|181
|296
|291
|.
|332
|232
|193
|-
|Tyr
|142
|173
|.
|194
|402
|357
|129
|87
|176
|369
|197
|340
|171
|392
|.
|362
|.
|360
|.
|303
|258
|-
|Trp
|137
|92
|17
|66
|63
|162
|.
|.
|65
|61
|239
|103
|54
|110
|.
|177
|110
|364
|281
|.
|142
|-
|Ex<sub>dest</sub>
|315
|311
|293
|192
|411
|321
|258
|225
|262
|305
|290
|255
|225
|314
|293
|307
|305
|294
|279
|172
|291
|} 
|}'''

    @classmethod
    def wiki_to_csv(cls,table):
        matrix = [[entry.lstrip().rstrip() for entry in row.split('|') if '|+' not in entry and '!' not in entry] for row in table.replace('{| class="wikitable"','').replace('|}','').replace('\n','').split('|-')]
        raise Error_working_here

    def __init__(self, table):
        self.data = {}
        matrix = [[entry.lstrip().rstrip() for entry in row.split('|') if '|+' not in entry and '!' not in entry] for row in table.replace('{| class="wikitable"','').replace('|}','').replace('\n','').split('|-')]
        from pprint import PrettyPrinter
        pprint = PrettyPrinter().pprint
        pprint(matrix)
        hheader = matrix[0]
        print(hheader)
        vheader = [row[0] for row in matrix[1:-1]]
        print(vheader)

        for i in range(1, len(matrix)):
            row = matrix[i]
            for j in range(1, len(row)):
                if row[j]:
                    self.data[hheader[i]][vheader[j]] = float(row[j])
                elif hheader[i] == vheader[j]:
                    self.data[hheader[i]][vheader[j]] = 0
                else:
                    self.data[hheader[i]][vheader[j]] = self.data[vheader[j]][hheader[i]]
        values = [self.data[k][l] for k in self.data for l in self.data[k]]
        self.ndata = {k: {l: sum(filter(lambda x: x < self.data[k][l], values)) for l in self.data[k]} for k in self.data}

if __name__ == '__main__':
    changedex = Changedex().fill()

    from pprint import PrettyPrinter
    pprint = PrettyPrinter().pprint

    pprint(changedex.to_dict())

    # {'A>G': 'IncreasedPolarity|Smaller|Flexibler', 'A>A': 'Identity', 'V>G': 'IncreasedPolarity|Smaller|Flexibler', 'V>A': 'Smaller', 'V>P': 'toAromatic|Smaller|Flexibler|IncreasedPolarity', 'V>S': 'IncreasedPolarity|Smaller|Flexibler', 'V>C': 'IncreasedPolarity|Smaller', 'I>G': 'IncreasedPolarity|Smaller|Flexibler', 'I>A': 'Smaller', 'I>P': 'toAromatic|Smaller|Flexibler|IncreasedPolarity', 'I>S': 'IncreasedPolarity|Smaller|Flexibler', 'I>C': 'IncreasedPolarity|Smaller', 'L>G': 'IncreasedPolarity|Smaller|Flexibler', 'L>A': 'Smaller', 'L>P': 'toAromatic|Smaller|Flexibler|IncreasedPolarity', 'L>S': 'IncreasedPolarity|Smaller|Flexibler', 'L>C': 'IncreasedPolarity|Smaller', 'M>G': 'IncreasedPolarity|Smaller|Flexibler', 'M>A': 'Smaller', 'M>P': 'toAromatic|Smaller|Flexibler|IncreasedPolarity', 'M>S': 'IncreasedPolarity|Smaller|Flexibler', 'M>C': 'IncreasedPolarity|Smaller', 'F>G': 'IncreasedPolarity|Smaller|Flexibler', 'F>A': 'Smaller', 'F>P': 'toAromatic|Smaller|Flexibler|IncreasedPolarity', 'F>S': 'IncreasedPolarity|Smaller|Flexibler', 'F>C': 'IncreasedPolarity|Smaller', 'Y>G': 'IncreasedPolarity|fromAromatic|Smaller|Flexibler', 'Y>A': 'fromAromatic|Smaller', 'Y>P': 'IncreasedPolarity|Smaller|Flexibler', 'Y>S': 'IncreasedPolarity|fromAromatic|Smaller|Flexibler', 'Y>C': 'IncreasedPolarity|fromAromatic|Smaller', 'W>G': 'IncreasedPolarity|fromAromatic|Smaller|Flexibler', 'W>A': 'fromAromatic|Smaller', 'W>P': 'IncreasedPolarity|Smaller|Flexibler', 'W>S': 'IncreasedPolarity|fromAromatic|Smaller|Flexibler', 'W>C': 'IncreasedPolarity|fromAromatic|Smaller', 'S>G': 'Smaller|Flexibler', 'S>A': 'Smaller|Rigider|IncreasedHydrophobicity', 'S>P': 'toAromatic|Smaller|Flexibler', 'S>S': 'Identity', 'S>C': 'Bigger|Smaller|Equal-sized|Rigider', 'T>G': 'Smaller|Flexibler', 'T>A': 'Smaller|IncreasedHydrophobicity', 'T>P': 'toAromatic|Smaller|Flexibler', 'T>S': 'Smaller|Flexibler', 'T>C': 'Smaller', 'N>G': 'Smaller|Flexibler', 'N>A': 'Smaller|IncreasedHydrophobicity', 'N>P': 'toAromatic|Smaller|Flexibler', 'N>S': 'Smaller|Flexibler', 'N>C': 'Smaller', 'Q>G': 'Smaller|Flexibler', 'Q>A': 'Smaller|IncreasedHydrophobicity', 'Q>P': 'toAromatic|Smaller|Flexibler', 'Q>S': 'Smaller|Flexibler', 'Q>C': 'Smaller', 'C>G': 'Smaller|Flexibler', 'C>A': 'Smaller|IncreasedHydrophobicity', 'C>P': 'toAromatic|Smaller|Flexibler', 'C>S': 'Bigger|Smaller|Equal-sized|Flexibler', 'C>C': 'Identity', 'G>G': 'Identity', 'P>G': 'fromAromatic|Smaller|Rigider', 'P>A': 'fromAromatic|Smaller|Rigider|IncreasedHydrophobicity', 'P>P': 'Identity', 'R>G': 'Charge-changed|Smaller|Flexibler', 'R>A': 'Charge-changed|Smaller', 'R>P': 'Charge-changed|toAromatic|Smaller|Flexibler', 'R>S': 'Charge-changed|Smaller|Flexibler', 'R>C': 'Charge-changed|Smaller', 'H>G': 'Charge-changed|fromAromatic|Smaller|Flexibler', 'H>A': 'Charge-changed|fromAromatic|Smaller', 'H>P': 'Charge-changed|Smaller|Flexibler', 'H>S': 'Charge-changed|fromAromatic|Smaller|Flexibler', 'H>C': 'Charge-changed|fromAromatic|Smaller', 'K>G': 'Charge-changed|Smaller|Flexibler', 'K>A': 'Charge-changed|Smaller', 'K>P': 'Charge-changed|toAromatic|Smaller|Flexibler', 'K>S': 'Charge-changed|Smaller|Flexibler', 'K>C': 'Charge-changed|Smaller', 'D>G': 'Charge-changed|Smaller|Flexibler', 'D>A': 'Charge-changed|Smaller', 'D>P': 'Charge-changed|toAromatic|Smaller|Flexibler', 'D>S': 'Charge-changed|Smaller|Flexibler', 'D>C': 'Charge-changed|Smaller', 'E>G': 'Charge-changed|Smaller|Flexibler', 'E>A': 'Charge-changed|Smaller', 'E>P': 'Charge-changed|toAromatic|Smaller|Flexibler', 'E>S': 'Charge-changed|Smaller|Flexibler', 'E>C': 'Charge-changed|Smaller', 'E>Q': 'Charge-changed|Equal-sized', 'D>N': 'Charge-changed|Equal-sized', 'L>D': 'Charge-changed|Equal-sized', 'L>N': 'IncreasedPolarity|Equal-sized', 'I>V': 'Smaller', 'Y>F': 'fromAromatic|Smaller', 'E>D': 'Smaller', 'E>N': 'Charge-changed|Smaller', 'Q>D': 'Charge-changed|Smaller', 'Q>N': 'Smaller', 'L>V': 'Smaller', 'F>H': 'Charge-changed|toAromatic|Smaller', 'Y>H': 'Charge-changed|Smaller', 'W>H': 'Charge-changed|Smaller', 'W>Y': 'Different-bulk', 'W>F': 'fromAromatic|Different-bulk', 'E>M': 'Charge-changed|Smaller', 'Q>M': 'Smaller|IncreasedHydrophobicity', 'F>V': 'Smaller', 'F>T': 'IncreasedPolarity|Smaller', 'F>L': 'Smaller', 'F>D': 'Charge-changed|Smaller', 'F>E': 'Charge-changed|Smaller', 'F>M': 'Smaller', 'Y>V': 'fromAromatic|Smaller', 'Y>T': 'IncreasedPolarity|fromAromatic|Smaller', 'Y>L': 'fromAromatic|Smaller', 'Y>D': 'Charge-changed|fromAromatic|Smaller', 'Y>E': 'Charge-changed|fromAromatic|Smaller', 'Y>M': 'fromAromatic|Smaller', 'W>V': 'fromAromatic|Smaller', 'W>T': 'IncreasedPolarity|fromAromatic|Smaller', 'W>L': 'fromAromatic|Smaller', 'W>D': 'Charge-changed|fromAromatic|Smaller', 'V>T': 'IncreasedPolarity|Equal-sized', 'R>M': 'Charge-changed|Smaller', 'R>E': 'Charge-changed|Smaller', 'R>D': 'Charge-changed|Smaller', 'K>M': 'Charge-changed|Smaller', 'K>E': 'Charge-changed|Smaller', 'K>D': 'Charge-changed|Smaller', 'R>K': 'Smaller', 'A>V': 'Bigger', 'A>I': 'Bigger', 'A>L': 'Bigger', 'A>M': 'Bigger', 'A>F': 'Bigger', 'A>Y': 'Bigger|toAromatic', 'A>W': 'Bigger|toAromatic', 'A>S': 'Bigger|Flexibler|IncreasedPolarity', 'A>T': 'Bigger|IncreasedPolarity', 'A>N': 'Bigger|IncreasedPolarity', 'A>Q': 'Bigger|IncreasedPolarity', 'A>C': 'Bigger|IncreasedPolarity', 'G>A': 'Bigger|Rigider|IncreasedHydrophobicity', 'A>P': 'Bigger|toAromatic|Flexibler|IncreasedPolarity', 'A>R': 'Charge-changed|Bigger', 'A>H': 'Charge-changed|Bigger|toAromatic', 'A>K': 'Charge-changed|Bigger', 'A>D': 'Charge-changed|Bigger', 'A>E': 'Charge-changed|Bigger', 'V>V': 'Identity', 'V>I': 'Bigger', 'V>L': 'Bigger', 'V>M': 'Different-bulk', 'V>F': 'Bigger', 'V>Y': 'Bigger|toAromatic', 'V>W': 'Bigger|toAromatic', 'S>V': 'Bigger|Rigider|IncreasedHydrophobicity', 'T>V': 'Equal-sized|IncreasedHydrophobicity', 'V>N': 'IncreasedPolarity|Different-bulk', 'V>Q': 'IncreasedPolarity|Different-bulk', 'C>V': 'Bigger|IncreasedHydrophobicity', 'G>V': 'Bigger|Rigider|IncreasedHydrophobicity', 'P>V': 'Bigger|fromAromatic|Rigider|IncreasedHydrophobicity', 'V>R': 'Charge-changed|Different-bulk', 'V>H': 'Charge-changed|toAromatic|Different-bulk', 'V>K': 'Charge-changed|Different-bulk', 'V>D': 'Charge-changed|Different-bulk', 'V>E': 'Charge-changed|Different-bulk', 'I>I': 'Identity', 'I>L': 'Different-bulk', 'I>M': 'Different-bulk', 'I>F': 'Different-bulk', 'I>Y': 'toAromatic|Different-bulk', 'I>W': 'toAromatic|Different-bulk', 'S>I': 'Bigger|Rigider|IncreasedHydrophobicity', 'I>T': 'IncreasedPolarity|Different-bulk', 'I>N': 'IncreasedPolarity|Different-bulk', 'I>Q': 'IncreasedPolarity|Different-bulk', 'C>I': 'Bigger|IncreasedHydrophobicity', 'G>I': 'Bigger|Rigider|IncreasedHydrophobicity', 'P>I': 'Bigger|fromAromatic|Rigider|IncreasedHydrophobicity', 'I>R': 'Charge-changed|Different-bulk', 'I>H': 'Charge-changed|toAromatic|Different-bulk', 'I>K': 'Charge-changed|Different-bulk', 'I>D': 'Charge-changed|Different-bulk', 'I>E': 'Charge-changed|Different-bulk', 'L>I': 'Different-bulk', 'L>L': 'Identity', 'L>M': 'Different-bulk', 'L>F': 'Bigger', 'L>Y': 'Bigger|toAromatic', 'L>W': 'Bigger|toAromatic', 'S>L': 'Bigger|Rigider|IncreasedHydrophobicity', 'L>T': 'IncreasedPolarity|Different-bulk', 'N>L': 'Equal-sized|IncreasedHydrophobicity', 'L>Q': 'IncreasedPolarity|Different-bulk', 'C>L': 'Bigger|IncreasedHydrophobicity', 'G>L': 'Bigger|Rigider|IncreasedHydrophobicity', 'P>L': 'Bigger|fromAromatic|Rigider|IncreasedHydrophobicity', 'L>R': 'Charge-changed|Different-bulk', 'L>H': 'Charge-changed|toAromatic|Different-bulk', 'L>K': 'Charge-changed|Different-bulk', 'D>L': 'Charge-changed|Equal-sized', 'L>E': 'Charge-changed|Different-bulk', 'M>V': 'Different-bulk', 'M>I': 'Different-bulk', 'M>L': 'Different-bulk', 'M>M': 'Identity', 'M>F': 'Bigger', 'M>Y': 'Bigger|toAromatic', 'M>W': 'toAromatic|Different-bulk', 'S>M': 'Bigger|Rigider|IncreasedHydrophobicity', 'M>T': 'IncreasedPolarity|Different-bulk', 'M>N': 'IncreasedPolarity|Different-bulk', 'M>Q': 'Bigger|IncreasedPolarity', 'C>M': 'Bigger|IncreasedHydrophobicity', 'G>M': 'Bigger|Rigider|IncreasedHydrophobicity', 'P>M': 'Bigger|fromAromatic|Rigider|IncreasedHydrophobicity', 'M>R': 'Charge-changed|Bigger', 'M>H': 'Charge-changed|toAromatic|Different-bulk', 'M>K': 'Charge-changed|Bigger', 'M>D': 'Charge-changed|Different-bulk', 'M>E': 'Charge-changed|Bigger', 'F>I': 'Different-bulk', 'F>F': 'Identity', 'F>Y': 'Bigger|toAromatic', 'F>W': 'toAromatic|Different-bulk', 'S>F': 'Bigger|Rigider|IncreasedHydrophobicity', 'T>F': 'Bigger|IncreasedHydrophobicity', 'F>N': 'IncreasedPolarity|Different-bulk', 'F>Q': 'IncreasedPolarity|Different-bulk', 'C>F': 'Bigger|IncreasedHydrophobicity', 'G>F': 'Bigger|Rigider|IncreasedHydrophobicity', 'P>F': 'Bigger|fromAromatic|Rigider|IncreasedHydrophobicity', 'F>R': 'Charge-changed|Different-bulk', 'H>F': 'Charge-changed|Bigger|fromAromatic', 'F>K': 'Charge-changed|Different-bulk', 'D>F': 'Charge-changed|Bigger', 'E>F': 'Charge-changed|Bigger', 'Y>I': 'fromAromatic|Different-bulk', 'Y>Y': 'Identity', 'Y>W': 'Different-bulk', 'S>Y': 'Bigger|toAromatic|Rigider|IncreasedHydrophobicity', 'T>Y': 'Bigger|toAromatic|IncreasedHydrophobicity', 'Y>N': 'IncreasedPolarity|fromAromatic|Different-bulk', 'Y>Q': 'IncreasedPolarity|fromAromatic|Different-bulk', 'C>Y': 'Bigger|toAromatic|IncreasedHydrophobicity', 'G>Y': 'Bigger|toAromatic|Rigider|IncreasedHydrophobicity', 'P>Y': 'Bigger|Rigider|IncreasedHydrophobicity', 'Y>R': 'Charge-changed|fromAromatic|Different-bulk', 'H>Y': 'Charge-changed|Bigger', 'Y>K': 'Charge-changed|fromAromatic|Different-bulk', 'D>Y': 'Charge-changed|Bigger|toAromatic', 'E>Y': 'Charge-changed|Bigger|toAromatic', 'W>I': 'fromAromatic|Different-bulk', 'W>M': 'fromAromatic|Different-bulk', 'W>W': 'Identity', 'S>W': 'Bigger|toAromatic|Rigider|IncreasedHydrophobicity', 'T>W': 'Bigger|toAromatic|IncreasedHydrophobicity', 'W>N': 'IncreasedPolarity|fromAromatic|Different-bulk', 'W>Q': 'IncreasedPolarity|fromAromatic|Different-bulk', 'C>W': 'Bigger|toAromatic|IncreasedHydrophobicity', 'G>W': 'Bigger|toAromatic|Rigider|IncreasedHydrophobicity', 'P>W': 'Bigger|Rigider|IncreasedHydrophobicity', 'W>R': 'Charge-changed|fromAromatic|Different-bulk', 'H>W': 'Charge-changed|Bigger', 'W>K': 'Charge-changed|fromAromatic|Different-bulk', 'D>W': 'Charge-changed|Bigger|toAromatic', 'W>E': 'Charge-changed|fromAromatic|Different-bulk', 'S>T': 'Bigger|Rigider', 'S>N': 'Bigger|Rigider', 'S>Q': 'Bigger|Rigider', 'G>S': 'Bigger|Rigider', 'P>S': 'Bigger|fromAromatic|Rigider', 'S>R': 'Charge-changed|Bigger|Rigider', 'S>H': 'Charge-changed|Bigger|toAromatic|Rigider', 'S>K': 'Charge-changed|Bigger|Rigider', 'S>D': 'Charge-changed|Bigger|Rigider', 'S>E': 'Charge-changed|Bigger|Rigider', 'T>I': 'Different-bulk|IncreasedHydrophobicity', 'T>L': 'Different-bulk|IncreasedHydrophobicity', 'T>M': 'Different-bulk|IncreasedHydrophobicity', 'T>T': 'Identity', 'T>N': 'Different-bulk', 'T>Q': 'Different-bulk', 'C>T': 'Bigger', 'G>T': 'Bigger|Rigider', 'P>T': 'Bigger|fromAromatic|Rigider', 'T>R': 'Charge-changed|Different-bulk', 'T>H': 'Charge-changed|toAromatic|Different-bulk', 'T>K': 'Charge-changed|Different-bulk', 'T>D': 'Charge-changed|Different-bulk', 'T>E': 'Charge-changed|Different-bulk', 'N>V': 'Different-bulk|IncreasedHydrophobicity', 'N>I': 'Different-bulk|IncreasedHydrophobicity', 'N>M': 'Different-bulk|IncreasedHydrophobicity', 'N>F': 'Different-bulk|IncreasedHydrophobicity', 'N>Y': 'toAromatic|Different-bulk|IncreasedHydrophobicity', 'N>W': 'toAromatic|Different-bulk|IncreasedHydrophobicity', 'N>T': 'Different-bulk', 'N>N': 'Identity', 'N>Q': 'Bigger', 'C>N': 'Bigger', 'G>N': 'Bigger|Rigider', 'P>N': 'Bigger|fromAromatic|Rigider', 'N>R': 'Charge-changed|Different-bulk', 'N>H': 'Charge-changed|toAromatic|Different-bulk', 'N>K': 'Charge-changed|Different-bulk', 'N>D': 'Charge-changed|Equal-sized', 'N>E': 'Charge-changed|Bigger', 'Q>V': 'Different-bulk|IncreasedHydrophobicity', 'Q>I': 'Different-bulk|IncreasedHydrophobicity', 'Q>L': 'Different-bulk|IncreasedHydrophobicity', 'Q>F': 'Different-bulk|IncreasedHydrophobicity', 'Q>Y': 'toAromatic|Different-bulk|IncreasedHydrophobicity', 'Q>W': 'toAromatic|Different-bulk|IncreasedHydrophobicity', 'Q>T': 'Different-bulk', 'Q>Q': 'Identity', 'C>Q': 'Bigger', 'G>Q': 'Bigger|Rigider', 'P>Q': 'Bigger|fromAromatic|Rigider', 'Q>R': 'Charge-changed|Different-bulk', 'Q>H': 'Charge-changed|toAromatic|Different-bulk', 'Q>K': 'Charge-changed|Different-bulk', 'D>Q': 'Charge-changed|Bigger', 'Q>E': 'Charge-changed|Equal-sized', 'G>C': 'Bigger|Rigider', 'P>C': 'Bigger|fromAromatic|Rigider', 'C>R': 'Charge-changed|Bigger', 'C>H': 'Charge-changed|Bigger|toAromatic', 'C>K': 'Charge-changed|Bigger', 'C>D': 'Charge-changed|Bigger', 'C>E': 'Charge-changed|Bigger', 'G>P': 'Bigger|toAromatic|Flexibler', 'G>R': 'Charge-changed|Bigger|Rigider', 'G>H': 'Charge-changed|Bigger|toAromatic|Rigider', 'G>K': 'Charge-changed|Bigger|Rigider', 'G>D': 'Charge-changed|Bigger|Rigider', 'G>E': 'Charge-changed|Bigger|Rigider', 'P>R': 'Charge-changed|Bigger|fromAromatic|Rigider', 'P>H': 'Charge-changed|Bigger|Rigider', 'P>K': 'Charge-changed|Bigger|fromAromatic|Rigider', 'P>D': 'Charge-changed|Bigger|fromAromatic|Rigider', 'P>E': 'Charge-changed|Bigger|fromAromatic|Rigider', 'R>V': 'Charge-changed|Different-bulk', 'R>I': 'Charge-changed|Different-bulk', 'R>L': 'Charge-changed|Different-bulk', 'R>F': 'Charge-changed|Different-bulk', 'R>Y': 'Charge-changed|toAromatic|Different-bulk', 'R>W': 'Charge-changed|toAromatic|Different-bulk', 'R>T': 'Charge-changed|Different-bulk', 'R>N': 'Charge-changed|Different-bulk', 'R>Q': 'Charge-changed|Different-bulk', 'R>R': 'Identity', 'R>H': 'toAromatic|Different-bulk', 'K>R': 'Bigger', 'D>R': 'Charge-changed|Bigger', 'E>R': 'Charge-changed|Bigger', 'H>V': 'Charge-changed|fromAromatic|Different-bulk', 'H>I': 'Charge-changed|fromAromatic|Different-bulk', 'H>L': 'Charge-changed|fromAromatic|Different-bulk', 'H>M': 'Charge-changed|fromAromatic|Different-bulk', 'H>T': 'Charge-changed|fromAromatic|Different-bulk', 'H>N': 'Charge-changed|fromAromatic|Different-bulk', 'H>Q': 'Charge-changed|fromAromatic|Different-bulk', 'H>R': 'fromAromatic|Different-bulk', 'H>H': 'Identity', 'H>K': 'fromAromatic|Different-bulk', 'H>D': 'Charge-changed|fromAromatic|Different-bulk', 'H>E': 'Charge-changed|fromAromatic|Different-bulk', 'K>V': 'Charge-changed|Different-bulk', 'K>I': 'Charge-changed|Different-bulk', 'K>L': 'Charge-changed|Different-bulk', 'K>F': 'Charge-changed|Different-bulk', 'K>Y': 'Charge-changed|toAromatic|Different-bulk', 'K>W': 'Charge-changed|toAromatic|Different-bulk', 'K>T': 'Charge-changed|Different-bulk', 'K>N': 'Charge-changed|Different-bulk', 'K>Q': 'Charge-changed|Different-bulk', 'K>H': 'toAromatic|Different-bulk', 'K>K': 'Identity', 'D>K': 'Charge-changed|Bigger', 'E>K': 'Charge-changed|Bigger', 'D>V': 'Charge-changed|Different-bulk', 'D>I': 'Charge-changed|Different-bulk', 'D>M': 'Charge-changed|Different-bulk', 'D>T': 'Charge-changed|Different-bulk', 'D>H': 'Charge-changed|toAromatic|Different-bulk', 'D>D': 'Identity', 'D>E': 'Bigger', 'E>V': 'Charge-changed|Different-bulk', 'E>I': 'Charge-changed|Different-bulk', 'E>L': 'Charge-changed|Different-bulk', 'E>W': 'Charge-changed|toAromatic|Different-bulk', 'E>T': 'Charge-changed|Different-bulk', 'E>H': 'Charge-changed|toAromatic|Different-bulk', 'E>E': 'Identity'}
